# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

import galsim
from galsim_test_helpers import *


@timer
def test_deltaFunction():
    """Test the generation of a Delta function profile
    """
    # Check construction with no arguments gives expected result
    delta = galsim.DeltaFunction()
    np.testing.assert_almost_equal(delta.flux, 1.0)
    check_basic(delta, "DeltaFunction")
    check_pickle(delta)

    # Check with default_params
    delta = galsim.DeltaFunction(flux=1, gsparams=default_params)
    np.testing.assert_almost_equal(delta.flux, 1.0)

    test_flux = 17.9
    delta = galsim.DeltaFunction(flux=test_flux)
    np.testing.assert_almost_equal(delta.flux, test_flux)
    check_basic(delta, "DeltaFunction")
    check_pickle(delta)

    gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
    delta2 = galsim.DeltaFunction(flux=test_flux, gsparams=gsp)
    assert delta2 != delta
    assert delta2 == delta.withGSParams(gsp)

    # Test operations with no-ops on DeltaFunction
    delta_shr = delta.shear(g1=0.3, g2=0.1)
    np.testing.assert_almost_equal(delta_shr.flux, test_flux)

    delta_dil = delta.dilate(2.0)
    np.testing.assert_almost_equal(delta_dil.flux, test_flux)

    delta_rot = delta.rotate(45 * galsim.radians)
    np.testing.assert_almost_equal(delta_rot.flux, test_flux)

    delta_tfm = delta.transform(dudx=1.25, dudy=0., dvdx=0., dvdy=0.8)
    np.testing.assert_almost_equal(delta_tfm.flux, test_flux)

    delta_shift = delta.shift(1.,2.)
    np.testing.assert_almost_equal(delta_shift.flux, test_flux)

    # These aren't no ops, since they do in fact alter the flux.
    delta_exp = delta.expand(2.0)
    np.testing.assert_almost_equal(delta_exp.flux, test_flux * 4)

    delta_mag = delta.magnify(2.0)
    np.testing.assert_almost_equal(delta_mag.flux, test_flux * 2)

    delta_tfm = delta.transform(dudx=1.4, dudy=0.2, dvdx=0.4, dvdy=1.2)
    np.testing.assert_almost_equal(delta_tfm.flux, test_flux * (1.4*1.2-0.2*0.4))

    # Test simple translation of DeltaFunction
    delta2 = delta.shift(1.,2.)
    offcen = galsim.PositionD(1, 2)
    np.testing.assert_equal(delta2.centroid, offcen)
    assert delta2.xValue(offcen) > 1.e10
    np.testing.assert_almost_equal(delta2.xValue(galsim.PositionD(0,0)), 0)

    # Test photon shooting.
    gauss = galsim.Gaussian(sigma = 1.0)
    delta_conv = galsim.Convolve(gauss,delta)
    myImg = galsim.ImageF()
    do_shoot(delta_conv,myImg,"Delta Function")

    # Test kvalues
    do_kvalue(delta_conv,myImg,"Delta Function")


@timer
def test_deltaFunction_properties():
    """Test some basic properties of the Delta function profile
    """
    test_flux = 17.9
    delta = galsim.DeltaFunction(flux=test_flux)
    # Check that we are centered on (0, 0)
    cen = galsim.PositionD(0, 0)
    np.testing.assert_equal(delta.centroid, cen)
    offcen = galsim.PositionD(1,1)
    # Check Fourier properties
    np.testing.assert_equal(delta.kValue(cen), (1+0j) * test_flux)
    np.testing.assert_equal(delta.kValue(offcen), (1+0j) * test_flux)
    import math
    assert delta.xValue(cen) > 1.e10
    np.testing.assert_almost_equal(delta.xValue(offcen), 0)
    assert delta.maxk > 1.e10
    assert delta.stepk > 1.e10
    # Check input flux vs output flux
    for inFlux in np.logspace(-2, 2, 10):
        delta = galsim.DeltaFunction(flux=inFlux)
        outFlux = delta.flux
        np.testing.assert_almost_equal(outFlux, inFlux)

@timer
def test_deltaFunction_flux_scaling():
    """Test flux scaling for Delta function.
    """
    # decimal point to go to for parameter value comparisons
    param_decimal = 12
    test_flux = 17.9

    # init with flux only (should be ok given last tests)
    obj = galsim.DeltaFunction(flux=test_flux)
    obj *= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __imul__.")
    obj = galsim.DeltaFunction(flux=test_flux)
    obj /= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __idiv__.")
    obj = galsim.DeltaFunction(flux=test_flux)
    obj2 = obj * 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (result).")
    obj = galsim.DeltaFunction(flux=test_flux)
    obj2 = 2. * obj
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (result).")
    obj = galsim.DeltaFunction(flux=test_flux)
    obj2 = obj / 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (result).")

    obj = galsim.DeltaFunction(flux=test_flux)
    obj2 = obj.withFlux(test_flux*2.)
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after obj.withFlux(flux).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after obj.withFlux(flux).")

@timer
def test_deltaFunction_convolution():
    """Test convolutions using the Delta function.
    """
    # Convolve two delta functions
    delta = galsim.DeltaFunction(flux=2.0)
    delta2 = galsim.DeltaFunction(flux=3.0)
    delta_delta = galsim.Convolve(delta,delta2)
    np.testing.assert_almost_equal(delta_delta.flux,6.0)

    # Test that no-ops on gaussians dont affect convolution
    gauss = galsim.Gaussian(sigma = 1.0)

    delta_shr = delta.shear(g1=0.3, g2=0.1)
    delta_conv = galsim.Convolve(gauss,delta_shr)
    np.testing.assert_almost_equal(delta_conv.flux,2.0)

    delta_dil = delta.dilate(2.0)
    delta_conv = galsim.Convolve(gauss,delta_dil)
    np.testing.assert_almost_equal(delta_conv.flux,2.0)

    delta_tfm = delta.transform(dudx=1.25, dudy=0.0, dvdx=0.0, dvdy=0.8)
    delta_conv = galsim.Convolve(gauss,delta_tfm)
    np.testing.assert_almost_equal(delta_conv.flux,2.0)

    delta_rot = delta.rotate(45 * galsim.radians)
    delta_conv = galsim.Convolve(gauss,delta_rot)
    np.testing.assert_almost_equal(delta_conv.flux,2.0)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'NFWHalo', 'Cosmology' ]

import numpy as np

from .position import PositionD
from .angle import arcsec
from . import integ
from . import utilities
from .errors import GalSimRangeError, GalSimIncompatibleValuesError

class Cosmology:
    """Basic cosmology calculations.

    Cosmology calculates expansion function E(a) and angular diameter distances Da(z) for a
    LambdaCDM universe.  Radiation is assumed to be zero and Dark Energy constant with w = -1 (no
    quintessence), but curvature is arbitrary.

    Based on Matthias Bartelmann's libastro.

    Parameters:
        omega_m:    Present day energy density of matter relative to critical density.
                    [default: 0.3]
        omega_lam:  Present day density of Dark Energy relative to critical density.
                    [default: 0.7]
    """
    def __init__(self, omega_m=0.3, omega_lam=0.7):
        # no quintessence, no radiation in this universe!
        self.omega_m = omega_m
        self.omega_lam = omega_lam
        self.omega_c = (1. - omega_m - omega_lam)
        #self.omega_r = 0

    def __repr__(self):
        return "galsim.Cosmology(omega_m=%r, omega_lam=%r)"%(self.omega_m, self.omega_lam)
    def __str__(self):
        return "galsim.Cosmology(%s,%s)"%(self.omega_m, self.omega_lam)
    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __ne__(self, other): return not self.__eq__(other)
    def __hash__(self): return hash(repr(self))

    def a(self, z):
        """Compute scale factor.

        Parameters:
            z:  Redshift
        """
        return 1./(1+z)

    def E(self, a):
        """Evaluates expansion function.

        Parameters:
            a:  Scale factor.
        """
        #return (self.omega_r*a**(-4) + self.omega_m*a**(-3) + self.omega_c*a**(-2) + \
        #        self.omega_lam)**0.5
        return (self.omega_m*a**(-3) + self.omega_c*a**(-2) + self.omega_lam)**0.5

    def __angKernel(self, x):
        """Integration kernel for angular diameter distance computation.
        """
        return self.E(x**-1)**-1

    def Da(self, z, z_ref=0):
        """Compute angular diameter distance between two redshifts in units of c/H0.

        In order to get the distance in Mpc/h, multiply by c/H0~3000.

        Parameters:
            z:      Redshift.
            z_ref:  Reference redshift, with z_ref <= z. [default: 0]
        """
        if isinstance(z, np.ndarray):
            da = np.zeros_like(z, dtype=float)
            for i in range(len(da)):
                da[i] = self.Da(z[i], z_ref)
            return da
        else:
            if z < 0:
                raise GalSimRangeError("Redshift z must be >= 0", z, 0.)
            if z < z_ref:
                raise GalSimRangeError("Redshift z must be >= the reference redshift", z, z_ref)

            d = integ.int1d(self.__angKernel, z_ref+1, z+1)
            # check for curvature
            rk = (abs(self.omega_c))**0.5
            if (rk*d > 0.01):
                if self.omega_c > 0:
                    d = np.sinh(rk*d)/rk
                if self.omega_c < 0:
                    d = np.sin(rk*d)/rk
            return d/(1+z)

class NFWHalo:
    """Class describing NFW halos.

    This class computes the lensing fields shear and convergence of a spherically symmetric NFW
    halo of given mass, concentration, redshift, assuming a particular cosmology.
    No mass-concentration relation is employed.

    Based on Matthias Bartelmann's libastro.

    The cosmology to use can be set either by providing a `Cosmology` instance as cosmo,
    or by providing omega_m and/or omega_lam.
    If only one of the latter is provided, the other is taken to be one minus that.
    If no cosmology parameters are set, a default `Cosmology` is constructed.

    Parameters:
        mass:       Mass defined using a spherical overdensity of 200 times the critical density
                    of the universe, in units of M_solar/h.
        conc:       Concentration parameter, i.e., ratio of virial radius to NFW scale radius.
        redshift:   Redshift of the halo.
        halo_pos:   `Position` of halo center (in arcsec). [default: PositionD(0,0)]
        omega_m:    Omega_matter to pass to `Cosmology` constructor. [default: None]
        omega_lam:  Omega_lambda to pass to `Cosmology` constructor. [default: None]
        cosmo:      A `Cosmology` instance. [default: None]
    """
    _req_params = { 'mass' : float , 'conc' : float , 'redshift' : float }
    _opt_params = { 'halo_pos' : PositionD , 'omega_m' : float , 'omega_lam' : float }

    def __init__(self, mass, conc, redshift, halo_pos=PositionD(0,0),
                 omega_m=None, omega_lam=None, cosmo=None):
        if omega_m is not None or omega_lam is not None:
            if cosmo is not None:
                raise GalSimIncompatibleValuesError(
                    "NFWHalo constructor received both cosmo and omega parameters",
                    cosmo=cosmo, omega_m=omega_m, omega_lam=omega_lam)
            if omega_m is None: omega_m = 1.-omega_lam
            if omega_lam is None: omega_lam = 1.-omega_m
            cosmo = Cosmology(omega_m=omega_m, omega_lam=omega_lam)
        elif cosmo is None:
            cosmo = Cosmology()
        elif not isinstance(cosmo,Cosmology):
            raise TypeError("Invalid cosmo parameter in NFWHalo constructor")

        # Make sure things are the right types.
        self.M = float(mass)
        self.c = float(conc)
        self.z = float(redshift)
        self.halo_pos = PositionD(halo_pos)
        self.cosmo = cosmo

        # calculate scale radius
        a = self.cosmo.a(self.z)
        # First we get the virial radius, which is defined for some spherical overdensity as
        # 3 M / [4 pi (r_vir)^3] = overdensity
        # Here we have overdensity = 200 * rhocrit, to determine R200. The factor of 1.63e-5 comes
        # from the following set of prefactors: (3 / (4 pi * 200 * rhocrit))^(1/3)
        # where rhocrit = 2.8e11 h^2 M_solar / Mpc^3.  The mass in the equation below is in
        # M_solar/h, which is how the final units are Mpc/h.
        R200 = 1.63e-5/(1+self.z) * (self.M * self.__omega(a)/self.__omega(1))**0.3333 # in Mpc/h
        self.rs = R200/self.c

        # convert scale radius in arcsec
        dl = self.cosmo.Da(self.z)*3000. # in Mpc/h
        scale = self.rs / dl
        arcsec2rad = 1./206265
        self.rs_arcsec = scale/arcsec2rad

    def __repr__(self):
        s = "galsim.NFWHalo(mass=%r, conc=%r, redshift=%r"%(self.M, self.c, self.z)
        if self.halo_pos != PositionD(0,0):
            s += ", halo_pos=%r"%self.halo_pos
        if self.cosmo != Cosmology():
            s += ", cosmo=%r"%self.cosmo
        s += ")"
        return s
    def __str__(self):
        return "galsim.NFWHalo(mass=%s, conc=%s, redshift=%s)"%(self.M, self.c, self.z)
    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __ne__(self, other): return not self.__eq__(other)
    def __hash__(self): return hash(repr(self))

    def __omega(self, a):
        """Matter density at scale factor a.
        """
        return self.cosmo.omega_m/(self.cosmo.E(a)**2 * a**3)

    def __farcth (self, x):
        """Numerical implementation of integral functions of a spherical NFW profile.

        All expressions are a function of ``x``, which is the radius r in units of the NFW scale
        radius, r_s.  For the derivation of these functions, see for example Wright & Brainerd
        (2000, ApJ, 534, 34).
        """
        out = np.zeros_like(x, dtype=float)

        # 3 cases: x > 1, x < 1, and |x-1| < 0.001
        mask = np.where(x < 0.999)[0]  # Equivalent but usually faster than `mask = (x < 0.999)`
        a = ((1.-x[mask])/(x[mask]+1.))**0.5
        out[mask] = 0.5*np.log((1.+a)/(1.-a))/(1-x[mask]**2)**0.5

        mask = np.where(x > 1.001)[0]
        a = ((x[mask]-1.)/(x[mask]+1.))**0.5
        out[mask] = np.arctan(a)/(x[mask]**2 - 1)**0.5

        # the approximation below has a maximum fractional error of 2.3e-7
        mask = np.where((x >= 0.999) & (x <= 1.001))[0]
        out[mask] = 5./6. - x[mask]/3.
        return out

    def __kappa(self, x, ks):
        """Calculate convergence of halo.

        Parameters:
            x:      Radial coordinate in units of rs (scale radius of halo), i.e., ``x=r/rs``.
            ks:     Lensing strength prefactor.
        """
        # convenience: call with single number
        if not isinstance(x, np.ndarray):
            return self.__kappa(np.array([x], dtype=float), np.array([ks], dtype=float))[0]
        out = np.zeros_like(x, dtype=float)

        # 3 cases: x > 1, x < 1, and |x-1| < 0.001
        mask = np.where(x < 0.999)[0]
        a = ((1 - x[mask])/(x[mask] + 1))**0.5
        out[mask] = 2*ks[mask]/(x[mask]**2 - 1) * (1 - np.log((1+a)/(1-a)) / (1-x[mask]**2)**0.5)

        mask = np.where(x > 1.001)[0]
        a = ((x[mask] - 1)/(x[mask] + 1))**0.5
        out[mask] = 2*ks[mask]/(x[mask]**2 - 1) * (1 - 2*np.arctan(a)/(x[mask]**2 - 1)**0.5)

        # the approximation below has a maximum fractional error of 7.4e-7
        mask = np.where((x >= 0.999) & (x <= 1.001))[0]
        out[mask] = ks[mask]*(22./15. - 0.8*x[mask])
        return out

    def __gamma(self, x, ks):
        """Calculate tangential shear of halo.

        Parameters:
            x:      Radial coordinate in units of rs (scale radius of halo), i.e., ``x=r/rs``.
            ks:     Lensing strength prefactor.
        """
        # convenience: call with single number
        if not isinstance(x, np.ndarray):
            return self.__gamma(np.array([x], dtype=float), np.array([ks], dtype=float))[0]
        out = np.zeros_like(x, dtype=float)

        mask = np.where(x > 0.01)[0]
        out[mask] = 4*ks[mask]*(np.log(x[mask]/2) + 2*self.__farcth(x[mask])) * \
            x[mask]**(-2) - self.__kappa(x[mask], ks[mask])

        # the approximation below has a maximum fractional error of 1.1e-7
        mask = np.where(x <= 0.01)[0]
        out[mask] = 4*ks[mask]*(0.25 + 0.125 * x[mask]**2 * (3.25 + 3.0*np.log(x[mask]/2)))
        return out

    def __ks(self, z_s):
        """Lensing strength of halo as function of source redshift.
        """
        # critical density and surface density
        rho_c = 2.7722e11
        Sigma_c = 5.5444e14
        # density contrast of halo at redshift z
        a = self.cosmo.a(self.z)
        ez = self.cosmo.E(a)
        d0 = 200./3 * self.c**3/(np.log(1+self.c) - (1.*self.c)/(1+self.c))
        rho_s = rho_c * ez**2 *d0

        # lensing weights: the only thing that depends on z_s
        # this does takes some time...
        dl = self.cosmo.Da(z_s, self.z) * self.cosmo.Da(self.z) / self.cosmo.Da(z_s)
        k_s = dl * self.rs * rho_s / Sigma_c
        return k_s

    def getShear(self, pos, z_s, units=arcsec, reduced=True):
        """Calculate (reduced) shear of halo at specified positions.

        Parameters:
            po:         Position(s) of the source(s), assumed to be post-lensing!
                        Valid ways to input this:

                        - single `Position` instance
                        - tuple of floats: (x,y)
                        - list/array of `Position` instances
                        - tuple of lists/arrays: ( xlist, ylist )

            z_s:        Source redshift(s).
            units:      Angular units of coordinates. [default: galsim.arcsec]
            reduced:    Whether returned shear(s) should be reduced shears. [default: True]

        Returns:
            the (possibly reduced) shears as a tuple (g1,g2)

        If the input ``pos`` is given a single position, (g1,g2) are the two shear components.
        If the input ``pos`` is given a list/array of positions, they are NumPy arrays.
        """
        pos_x, pos_y = utilities._convertPositions(pos, units, 'getShear')
        return self._getShear(pos_x, pos_y, z_s, reduced)

    def _getShear(self, pos_x, pos_y, z_s, reduced=True):
        """Equivalent to `getShear`, but without some sanity checks and the positions must be
        given as ``pos_x``, ``pos_y`` in arcsec.

        Parameters:
            pos_x:      x position in arcsec (either a scalar or a numpy array)
            pos_y:      y position in arcsec (either a scalar or a numpy array)
            z_s:        Source redshift(s).
            reduced:    Whether returned shear(s) should be reduced shears. [default: True]

        Returns:
            the (possibly reduced) shears as a tuple (g1,g2) (either scalars or numpy arrays)
        """
        r = ((pos_x - self.halo_pos.x)**2 + (pos_y - self.halo_pos.y)**2)**0.5/self.rs_arcsec
        # compute strength of lensing fields
        ks = self.__ks(z_s)
        if not isinstance(z_s, np.ndarray):
            ks = ks*np.ones_like(r)
        g = self.__gamma(r, ks)

        # convert to observable = reduced shear
        if reduced:
            kappa = self.__kappa(r, ks)
            g /= 1 - kappa

        # pure tangential shear, no cross component
        dx = pos_x - self.halo_pos.x
        dy = pos_y - self.halo_pos.y
        drsq = dx*dx+dy*dy
        # Avoid division by 0
        cos2phi = np.divide(dx*dx-dy*dy, drsq, where=(drsq != 0.))
        sin2phi = np.divide(2*dx*dy, drsq, where=(drsq != 0.))
        g1 = -g*cos2phi
        g2 = -g*sin2phi
        return g1, g2


    def getConvergence(self, pos, z_s, units=arcsec):
        """Calculate convergence of halo at specified positions.

        Parameters:
            pos:        Position(s) of the source(s), assumed to be post-lensing!
                        Valid ways to input this:

                        - single `Position` instance
                        - tuple of floats: (x,y)
                        - list/array of `Position` instances
                        - tuple of lists/arrays: ( xlist, ylist )

            z_s:        Source redshift(s).
            unit:       Angular units of coordinates. [default: galsim.arcsec]

        Returns:
            the convergence, kappa

        If the input ``pos`` is given a single position, kappa is the convergence value.
        If the input ``pos`` is given a list/array of positions, kappa is a NumPy array.
        """

        # Convert to numpy arrays for internal usage:
        pos_x, pos_y = utilities._convertPositions(pos, units, 'getKappa')
        return self._getConvergence(pos_x, pos_y, z_s)

    def _getConvergence(self, pos_x, pos_y, z_s):
        """Equivalent to `getConvergence`, but without some sanity checks and the positions must be
        given as ``pos_x``, ``pos_y`` in arcsec.

        Parameters:
            pos_x:      x position in arcsec (either a scalar or a numpy array)
            pos_y:      y position in arcsec (either a scalar or a numpy array)
            z_s:        Source redshift(s).

        Returns:
            the convergence as either a scalar or a numpy array
        """
        r = ((pos_x - self.halo_pos.x)**2 + (pos_y - self.halo_pos.y)**2)**0.5/self.rs_arcsec
        # compute strength of lensing fields
        ks = self.__ks(z_s)
        if not isinstance(z_s, np.ndarray):
            ks = ks*np.ones_like(r)
        kappa = self.__kappa(r, ks)
        return kappa

    def getMagnification(self, pos, z_s, units=arcsec):
        """Calculate magnification of halo at specified positions.

        Parameters:
            pos:        Position(s) of the source(s), assumed to be post-lensing!
                        Valid ways to input this:

                        - single `Position` instance
                        - tuple of floats: (x,y)
                        - list/array of `Position` instances
                        - tuple of lists/arrays: ( xlist, ylist )

            z_s:        Source redshift(s).
            units:      Angular units of coordinates. [default: galsim.arcsec]

        Returns:
            the magnification mu

        If the input ``pos`` is given a single position, mu is the magnification value.
        If the input ``pos`` is given a list/array of positions, mu is a NumPy array.
        """
        # Convert to numpy arrays for internal usage:
        pos_x, pos_y = utilities._convertPositions(pos, units, 'getMagnification')
        return self._getMagnification(pos_x, pos_y, z_s)

    def _getMagnification(self, pos_x, pos_y, z_s):
        """Equivalent to `getMagnification`, but without some sanity checks and the positions must
        be given as ``pos_x``, ``pos_y`` in arcsec.

        Parameters:
            pos_x:      x position in arcsec (either a scalar or a numpy array)
            pos_y:      y position in arcsec (either a scalar or a numpy array)
            z_s:        Source redshift(s).

        Returns:
            the magnification as either a scalar or a numpy array
        """
        r = ((pos_x - self.halo_pos.x)**2 + (pos_y - self.halo_pos.y)**2)**0.5/self.rs_arcsec
        # compute strength of lensing fields
        ks = self.__ks(z_s)
        if not isinstance(z_s, np.ndarray):
            ks = ks*np.ones_like(r)
        g = self.__gamma(r, ks)
        kappa = self.__kappa(r, ks)

        mu = 1. / ( (1.-kappa)**2 - g**2 )
        return mu

    def getLensing(self, pos, z_s, units=arcsec):
        """Calculate lensing shear and magnification of halo at specified positions.

        Parameters:
            pos:        Position(s) of the source(s), assumed to be post-lensing!
                        Valid ways to input this:

                        - single `Position` instance
                        - tuple of floats: (x,y)
                        - list/array of `Position` instances
                        - tuple of lists/arrays: ( xlist, ylist )

            z_s:        Source redshift(s).
            units:      Angular units of coordinates. [default: galsim.arcsec]

        Returns:
            the reduced shears and magnifications as a tuple (g1,g2,mu)

        If the input ``pos`` is given a single position, the return values are the shear and
        magnification values at that position.
        If the input ``pos`` is given a list/array of positions, they are NumPy arrays.
        """
        # Convert to numpy arrays for internal usage:
        pos_x, pos_y = utilities._convertPositions(pos, units, 'getLensing')
        return self._getLensing(pos_x, pos_y, z_s)

    def _getLensing(self, pos_x, pos_y, z_s):
        """Equivalent to `getLensing`, but without some sanity checks and the positions must
        be given as ``pos_x``, ``pos_y`` in arcsec.

        Parameters:
            pos_x:      x position in arcsec (either a scalar or a numpy array)
            pos_y:      y position in arcsec (either a scalar or a numpy array)
            z_s:        Source redshift(s).

        Returns:
            the reduced shears and magnifications as a tuple (g1,g2,mu) (each being
            either a scalar or a numpy array)
        """
        r = ((pos_x - self.halo_pos.x)**2 + (pos_y - self.halo_pos.y)**2)**0.5/self.rs_arcsec
        # compute strength of lensing fields
        ks = self.__ks(z_s)
        if not isinstance(z_s, np.ndarray):
            ks = ks*np.ones_like(r)
        g = self.__gamma(r, ks)
        kappa = self.__kappa(r, ks)

        mu = 1. / ( (1.-kappa)**2 - g**2 )
        g /= 1 - kappa
        # Get the tangential shear (no x component)
        dx = pos_x - self.halo_pos.x
        dy = pos_y - self.halo_pos.y
        drsq = dx*dx+dy*dy
        # Avoid division by 0
        cos2phi = np.divide(dx*dx-dy*dy, drsq, where=(drsq != 0.))
        sin2phi = np.divide(2*dx*dy, drsq, where=(drsq != 0.))
        g1 = -g*cos2phi
        g2 = -g*sin2phi
        return g1, g2, mu



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

from . import _galsim
from .errors import GalSimError, GalSimValueError, convert_cpp_errors
from ._utilities import LRU_Cache


def int1d(func, min, max, rel_err=1.e-6, abs_err=1.e-12):
    """Integrate a 1-dimensional function from min to max.

    Example usage::

        >>> def func(x): return x**2
        >>> galsim.integ.int1d(func, 0, 1)
        0.33333333333333337
        >>> galsim.integ.int1d(func, 0, 2)
        2.666666666666667
        >>> galsim.integ.int1d(func, -1, 1)
        0.66666666666666674

    .. note::

        This uses an adaptive Gauss-Kronrod-Patterson method for doing the integration.

        cf. https://www.jstor.org/stable/2004583

        If one or both endpoints are infinite, it will automatically use an appropriate
        transformation to turn it into a finite integral.

    Parameters:
        func:       The function to be integrated.  y = func(x) should be valid.
        min:        The lower end of the integration bounds (anything < -1.e10 is treated as
                    negative infinity).
        max:        The upper end of the integration bounds (anything > 1.e10 is treated as positive
                    infinity).
        rel_err:    The desired relative error [default: 1.e-6]
        abs_err:    The desired absolute error [default: 1.e-12]

    Returns:
        the value of the integral.
    """
    min = float(min)
    max = float(max)
    rel_err = float(rel_err)
    abs_err = float(abs_err)
    with convert_cpp_errors():
        success, result = _galsim.PyInt1d(func,min,max,rel_err,abs_err)
    if success:
        return result
    else:
        raise GalSimError(result)

def hankel(func, k, nu=0, rmax=None, rel_err=1.e-6, abs_err=1.e-12):
    r"""Perform an order nu Hankel transform of the given function f(r) at a specific k value.

    .. math::

        F(k) = \int_0^\infty f(r) J_\nu(k r) r dr

    .. note::

        For non-truncated Hankel integrals, this uses the method outlined in Ogata, 2005:
        http://www.kurims.kyoto-u.ac.jp/~prims/pdf/41-4/41-4-40.pdf

        For truncated integrals (and k=0), it uses the same adaptive Gauss-Kronrod-Patterson
        method used for `int1d`.

    Parameters:

        func:       The function f(r)
        k:          (float or numpy array) The value(s) of k for which to calculate F(k).
        nu:         The order of the Bessel function to use for the transform. [default: 0]
        rmax:       An optional truncation radius at which to have f(r) drop to 0. [default: None]
        rel_err:    The desired relative accuracy [default: 1.e-6]
        abs_err:    The desired absolute accuracy [default: 1.e-12]

    Returns:

        An estimate of F(k)
    """
    rel_err = float(rel_err)
    abs_err = float(abs_err)
    nu = float(nu)
    rmax = float(rmax) if rmax is not None else 0.

    k = np.ascontiguousarray(k, dtype=float)
    res = np.empty_like(k, dtype=float)
    N = 1 if k.shape == () else len(k)

    if np.any(k < 0):
        raise GalSimValueError("k must be >= 0",k)
    if nu < 0:
        raise GalSimValueError("nu must be >= 0",k)
    _k = k.__array_interface__['data'][0]
    _res = res.__array_interface__['data'][0]
    with convert_cpp_errors():
        _galsim.PyHankel(func, _k, _res, N, nu, rmax, rel_err, abs_err)
    return res

class IntegrationRule:
    """A class that can be used to integrate something more complicated than a normal
    scalar function.

    In GalSim, we use it to do the integration of chromatic images over a bandpass.
    Typically f is some kind of draw function, xs are the wavelengths, and w is the
    bandpass throughput.  But this class is abstracted away from all of that and can be used
    for anything where the function returns something complicated, but which can be added
    together to compute the quadrature.

    Specifically the return value from f must be closed under both addition and multiplication
    by a scalar (a float value).
    """
    def __init__(self):
        pass

    def __call__(self, f, xs, w=None):
        """Calculate the integral int(f(x) w(x) dx) using the appropriate Rule.

        Parameters:
            f:      Function to integrate.
            xs:     Locations at which to evaluate f.
            w:      Weight function if desired [default: None]

        Returns:
            The approximation to the integral.
        """
        gs = self.calculateWeights(xs, w)
        return sum(g * f(x) for g,x in zip(gs, xs))

class MidptRule(IntegrationRule):
    """Midpoint rule for integration.
    """
    def calculateWeights(self, xs, w):
        """Calculate the apporpriate weights for the midpoint rule integration

        Parameters:
            xs:     Locations at which to evaluate f.
            w:      Weight function if desired [default: None]

        Returns:
            The net weights to use at each location.
        """
        if len(xs) < 2:
            raise GalSimValueError("Not enough points for midptRule integration", xs)
        xs = np.asarray(xs)
        gs = np.empty_like(xs)
        gs[0] = (xs[1] - xs[0])
        gs[1:-1] = 0.5 * (xs[2:] - xs[:-2])
        gs[-1] = (xs[-1] - xs[-2])
        if w is not None:
            gs *= w(xs)
        return gs

class TrapzRule(IntegrationRule):
    """Trapezoidal rule for integration.
    """
    def calculateWeights(self, xs, w):
        """Calculate the apporpriate weights for the trapezoidal rule integration

        Parameters:
            xs:     Locations at which to evaluate f.
            w:      Weight function if desired [default: None]

        Returns:
            The net weights to use at each location.
        """
        if len(xs) < 2:
            raise GalSimValueError("Not enough points for trapzRule integration", xs)
        xs = np.asarray(xs)
        gs = np.empty_like(xs)
        gs[0] = 0.5 * (xs[1] - xs[0])
        gs[1:-1] = 0.5 * (xs[2:] - xs[:-2])
        gs[-1] = 0.5 * (xs[-1] - xs[-2])
        if w is not None:
            gs *= w(xs)
        return gs

class QuadRule(IntegrationRule):
    """Quadratic rule for integration

    This models both f and w as linear between the evaluation points, so the product is
    quadratic.
    """
    def calculateWeights(self, xs, w):
        """Calculate the apporpriate weights for the quadratic rule integration

        Parameters:
            xs:     Locations at which to evaluate f.
            w:      Weight function if desired [default: None]

        Returns:
            The net weights to use at each location.
        """
        if len(xs) < 2:
            raise GalSimValueError("Not enough points for quadRule integration", xs)
        if w is None:
            return TrapzRule().calculateWeights(xs,w)
        xs = np.asarray(xs)
        ws = w(xs)
        gs = np.empty_like(xs)
        gs[0] = (xs[1] - xs[0]) * (2*ws[0] + ws[1])
        gs[1:-1] = (xs[1:-1] - xs[:-2]) * (ws[:-2] + 2*ws[1:-1])
        gs[1:-1] += (xs[2:] - xs[1:-1]) * (2*ws[1:-1] + ws[2:])
        gs[-1] = (xs[-1] - xs[-2]) * (ws[-2] + 2*ws[-1])
        gs /= 6.
        return gs

# To ease backwards compatibility, these are an instantiated object of the above classes
midptRule = MidptRule()  #: For convenience, an instance of `MidptRule`
trapzRule = TrapzRule()  #: For convenience, an instance of `TrapzRule`
quadRule = QuadRule()    #: For convenience, an instance of `QuadRule`


class ImageIntegrator:
    """A base class for integrators used by `ChromaticObject` to integrate the drawn images
    over wavelengthh using a `Bandpass` as a weight function.
    """
    def __init__(self):
        raise NotImplementedError("Must instantiate subclass of ImageIntegrator")
    # subclasses must define
    # 1) a method `.calculateWaves(bandpass)` which will return the wavelengths at which to
    #    evaluate the integrand
    # 2) an function attribute `.rule` which takes an integrand function as its first
    #    argument, and a list of evaluation wavelengths as its second argument, and returns
    #    an approximation to the integral.  (E.g., the function midptRule above)

    def __call__(self, evaluateAtWavelength, bandpass, image, drawImageKwargs, doK=False):
        """
        Parameters:
            evaluateAtWavelength:   Function that returns a monochromatic surface brightness
                                    profile as a function of wavelength.
            bandpass:               `Bandpass` object representing the filter being imaged through.
            image:                  `Image` used to set size and scale of output
            drawImageKwargs:        dict with other kwargs to send to `ChromaticObject.drawImage`
                                    function.
            doK:                    Integrate up results of `ChromaticObject.drawKImage` instead of
                                    results of `ChromaticObject.drawImage`.  [default: False]

        Returns:
            the result of integral as an `Image`
        """
        waves = self.calculateWaves(bandpass)
        self.last_n_eval = len(waves)
        drawImageKwargs.pop('add_to_image', None) # Make sure add_to_image isn't in kwargs

        def integrand(w):
            prof = evaluateAtWavelength(w)
            if not doK:
                return prof.drawImage(image=image.copy(), **drawImageKwargs)
            else:
                return prof.drawKImage(image=image.copy(), **drawImageKwargs)
        return self.rule(integrand, waves, bandpass)


class SampleIntegrator(ImageIntegrator):
    """Create a chromatic surface brightness profile integrator, which will integrate over
    wavelength using a `Bandpass` as a weight function.

    This integrator will evaluate the integrand only at the wavelengths in ``bandpass.wave_list``.
    See ContinuousIntegrator for an integrator that evaluates the integrand at a given number of
    points equally spaced apart.

    Parameters:
        rule:       Which integration rule to apply to the wavelength and monochromatic surface
                    brightness samples.  Options include:

                    - galsim.integ.midptRule: Use the midpoint integration rule
                    - galsim.integ.trapzRule: Use the trapezoidal integration rule
                    - galsim.integ.quadRule: Use the quadratic integration rule

    """
    def __init__(self, rule):
        self.rule = rule

    def calculateWaves(self, bandpass):
        return bandpass.wave_list


class ContinuousIntegrator(ImageIntegrator):
    """Create a chromatic surface brightness profile integrator, which will integrate over
    wavelength using a `Bandpass` as a weight function.

    This integrator will evaluate the integrand at a given number ``N`` of equally spaced
    wavelengths over the interval defined by bandpass.blue_limit and bandpass.red_limit.  See
    SampleIntegrator for an integrator that only evaluates the integrand at the predefined set of
    wavelengths in ``bandpass.wave_list``.

    Parameters:
        rule:           Which integration rule to apply to the wavelength and monochromatic
                        surface brightness samples.  Options include:

                        - galsim.integ.midptRule: Use the midpoint integration rule
                        - galsim.integ.trapzRule: Use the trapezoidal integration rule
                        - galsim.integ.quadRule: Use the quadratic integration rule

        N:              Number of equally-wavelength-spaced monochromatic surface brightness
                        samples to evaluate. [default: 250]
        use_endpoints:  Whether to sample the endpoints ``bandpass.blue_limit`` and
                        ``bandpass.red_limit``.  This should probably be True for a rule like
                        numpy.trapz, which explicitly samples the integration limits.  For a
                        rule like the midpoint rule, however, the integration limits are not
                        generally sampled, (only the midpoint between each integration limit and
                        its nearest interior point is sampled), thus ``use_endpoints`` should be
                        set to False in this case.  [default: True]
    """
    def __init__(self, rule, N=250, use_endpoints=True):
        self.rule = rule
        self.N = N
        self.use_endpoints = use_endpoints

    def calculateWaves(self, bandpass):
        h = (bandpass.red_limit*1.0 - bandpass.blue_limit)/self.N
        if self.use_endpoints:
            return [bandpass.blue_limit + h * i for i in range(self.N+1)]
        else:
            return [bandpass.blue_limit + h * (i+0.5) for i in range(self.N)]


_leggauss = LRU_Cache(np.polynomial.legendre.leggauss)


def gq_annulus_points(r_outer, r_inner, n_rings, n_spokes):
    r""" Generate points and weights for Gaussian quadrature over an annulus.

    Sample points are generated on a grid of rings (constant radius) and spokes
    (constant azimuth).  The sample points with the weights can be used to
    approximate an integral over an annulus as

    .. math::
        \Int_annulus f(x, y) dx dy = \Sum_{x, y, w} f(x, y) * w

    To integrate a nth order 2d polynomial over an annulus exactly requires
        n_rings = n//2+1
        n_spokes = n+1

    References:
      - G. W. Forbes,
        "Optical system assessment for design: numerical ray tracing in the Gaussian pupil,"
        J. Opt. Soc. Am. A 5, 1943-1956 (1988)

      - Brian J. Bauman, Hong Xiao,
        "Gaussian quadrature for optical design with noncircular pupils and fields, and broad wavelength range,"
        Proc. SPIE 7652, International Optical Design Conference 2010, 76522S (9 September 2010); https://doi.org/10.1117/12.872773

    Parameters:
        r_outer: Outer radius of annulus
        r_inner: Inner radius of annulus
        n_rings: Number of sample rings to use in the Gaussian quadrature
        n_spokes: Number of sample spokes to use in the Gaussian quadrature

    Returns:
        x, y: Cartesian coordinates of the sample points
        weights: Weights to use for the Gaussian quadrature

    """
    Li, w = _leggauss(n_rings)
    eps = r_inner/r_outer
    area = np.pi*(r_outer**2 - r_inner**2)
    rings = np.sqrt(eps**2 + (1+Li)*(1-eps**2)/2)*r_outer
    spokes = np.linspace(0, 2*np.pi, n_spokes, endpoint=False)
    weights = w*area/(2*n_spokes)
    rings, spokes = np.meshgrid(rings, spokes)
    weights = np.broadcast_to(weights, rings.shape)
    rings = rings.ravel()
    spokes = spokes.ravel()
    weights = weights.ravel()
    x = rings*np.cos(spokes)
    y = rings*np.sin(spokes)
    return x, y, weights


def gq_annulus(f, r_outer, r_inner, n_rings, n_spokes):
    """ Integrate a function over an annular region, using Gaussian quadrature
    over an annulus.

    References:
      - G. W. Forbes,
        "Optical system assessment for design: numerical ray tracing in the Gaussian pupil,"
        J. Opt. Soc. Am. A 5, 1943-1956 (1988)

      - Brian J. Bauman, Hong Xiao,
        "Gaussian quadrature for optical design with noncircular pupils and fields, and broad wavelength range,"
        Proc. SPIE 7652, International Optical Design Conference 2010, 76522S (9 September 2010); https://doi.org/10.1117/12.872773

    The annulus is sampled on a grid of rings and spokes.
    To integrate a nth order 2d polynomial exactly, this technique requires
        n_rings = n//2+1
        n_spokes = n+1

    Parameters:
        f: Function to integrate
        r_inner: Inner radius of annulus
        r_outer: Outer radius of annulus
        n_rings: Number of sample rings to use in the Gaussian quadrature
        n_spokes: Number of sample spokes to use in the Gaussian quadrature

    Returns:
        The integral of f over the annulus
    """
    x, y, weights = gq_annulus_points(r_outer, r_inner, n_rings, n_spokes)
    val = np.array([f(x_, y_) for x_, y_ in zip(x, y)])
    return np.sum(val*weights)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import time

from .output import OutputBuilder, RegisterOutputType
from .util import CopyConfig
from .image import BuildImages, BuildImage, GetNObjForImage
from .input import ProcessInputNObjects
from .value import ParseValue, CheckAllParams
from ..errors import GalSimConfigError
from ..fits import writeCube

class DataCubeBuilder(OutputBuilder):
    """Builder class for constructing and writing DataCube output types.
    """

    def buildImages(self, config, base, file_num, image_num, obj_num, ignore, logger):
        """Build the images

        A point of attention for DataCubes is that they must all be the same size.
        This function builds the first image alone, finds out its size and then forces
        all subsequent images to be the same size.

        Parameters:
            config:         The configuration dict for the output field.
            base:           The base configuration dict.
            file_num:       The current file_num.
            image_num:      The current image_num.
            obj_num:        The current obj_num.
            ignore:         A list of parameters that are allowed to be in config that we can
                            ignore here.  i.e. it won't be an error if they are present.
            logger:         If given, a logger object to log progress.

        Returns:
            a list of the images built
        """
        nimages = self.getNImages(config, base, file_num, logger=logger)

        # The above call sets up a default nimages if appropriate.  Now, check that there are no
        # invalid parameters in the config dict.
        req = { 'nimages' : int }
        ignore += [ 'file_name', 'dir', 'nfiles' ]
        CheckAllParams(config, ignore=ignore, req=req)

        # All images need to be the same size for a data cube.
        # Enforce this by building the first image outside the below loop and setting
        # config['image_force_xsize'] and config['image_force_ysize'] to be the size of the first
        # image.
        t1 = time.time()
        base1 = CopyConfig(base)
        image0 = BuildImage(base1, image_num, obj_num, logger=logger)
        t2 = time.time()
        # Note: numpy shape is y,x
        ys, xs = image0.array.shape
        logger.info('Image %d: size = %d x %d, time = %f sec', image_num, xs, ys, t2-t1)

        # Note: numpy shape is y,x
        image_ysize, image_xsize = image0.array.shape
        base['image_force_xsize'] = image_xsize
        base['image_force_ysize'] = image_ysize

        images = [ image0 ]

        if nimages > 1:
            obj_num += GetNObjForImage(base, image_num, logger=logger)
            images += BuildImages(nimages-1, base, logger=logger,
                                  image_num=image_num+1, obj_num=obj_num)

        return images

    def getNImages(self, config, base, file_num, logger=None):
        """Returns the number of images to be built.

        Parameters:
            config:         The configuration dict for the output field.
            base:           The base configuration dict.
            file_num:       The current file number.
            logger:         If given, a logger object to log progress.

        Returns:
            the number of images to build.
        """
        # Allow nimages to be automatic based on input catalog if image type is Single
        if ( 'nimages' not in config and
            ( 'image' not in base or 'type' not in base['image'] or
            base['image']['type'] == 'Single' ) ):
            nimages = ProcessInputNObjects(base)
            if nimages:
                config['nimages'] = nimages
        if 'nimages' not in config:
            raise GalSimConfigError(
                "Attribute output.nimages is required for output.type = MultiFits")
        return ParseValue(config,'nimages',base,int)[0]

    def writeFile(self, data, file_name, config, base, logger):
        """Write the data to a file.

        Parameters:
            data:           The data to write.  Usually a list of images returned by
                            buildImages, but possibly with extra HDUs tacked onto the end
                            from the extra output items.
            file_name:      The file_name to write to.
            config:         The configuration dict for the output field.
            base:           The base configuration dict.
            logger:         If given, a logger object to log progress.
        """
        writeCube(data,file_name)

    def canAddHdus(self):
        """Returns whether it is permissible to add extra HDUs to the end of the data list.

        False for DataCube.
        """
        return False


# Register this as a valid output type
RegisterOutputType('DataCube', DataCubeBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Transform', 'Transformation', '_Transform', ]

import numpy as np
import math
import cmath
import copy

from . import _galsim
from .gsobject import GSObject
from .gsparams import GSParams
from .utilities import lazy_property, WeakMethod
from .position import PositionD, _PositionD, parse_pos_args
from .errors import GalSimError
from .sum import Sum
from .sed import SED
from .wcs import JacobianWCS
from .shear import Shear, _Shear
from .angle import Angle
from . import convolve
from . import chromatic as chrom

def Transform(obj, jac=None, offset=(0.,0.), flux_ratio=1., gsparams=None,
              propagate_gsparams=True):
    """A function for transforming either a `GSObject` or `ChromaticObject`.

    This function will inspect its input argument to decide if a `Transformation` object or a
    `ChromaticTransformation` object is required to represent the resulting transformed object.

    Note: the name of the flux_ratio parameter is technically wrong here if the jacobian has a
    non-unit determinant, since that would also scale the flux.  The flux_ratio parameter actually
    only refers to an overall amplitude ratio for the surface brightness profile.  The total
    flux scaling is actually ``|det(jac)| * flux_ratio``.

    Parameters:
        obj:                The object to be transformed.
        jac:                A list or tuple ( dudx, dudy, dvdx, dvdy ) describing the Jacobian
                            of the transformation.  Use None to indicate that the Jacobian is the
                            2x2 unit matrix.  [default: None]
        offset:             A galsim.PositionD or tuple giving the offset by which to shift the
                            profile. [default: (0.,0.)]
        flux_ratio:         A factor by which to multiply the surface brightness of the object.
                            (Technically, not necessarily the flux.  See above.) [default: 1]
        gsparams:           An optional `GSParams` argument. [default: None]
        propagate_gsparams: Whether to propagate gsparams to the transformed object.  This
                            is normally a good idea, but there may be use cases where one
                            would not want to do this. [default: True]

    Returns:
        a `Transformation` or `ChromaticTransformation` instance as appropriate.
    """
    # Usual cases are a bit simpler to handle, so check for them first.
    if (isinstance(obj, (GSObject, chrom.SimpleChromaticTransformation))
        and not hasattr(jac,'__call__') and not hasattr(offset,'__call__')):

        if isinstance(obj, GSObject) :
            if not hasattr(flux_ratio,'__call__'):
                # This is the usual achromatic case, so check for this first.
                return Transformation(obj, jac, offset, flux_ratio, gsparams, propagate_gsparams)

            elif jac is None and offset == (0.,0.):
                # This is obj * SED.  Return a SimpleChromaticTransformation.
                if not isinstance(flux_ratio, SED):
                    flux_ratio = SED(flux_ratio, 'nm', '1')
                return chrom.SimpleChromaticTransformation(obj, flux_ratio, gsparams=gsparams,
                                                           propagate_gsparams=propagate_gsparams)

            else:
                # Too complicated for simple.
                return chrom.ChromaticTransformation(obj, jac, offset, flux_ratio,
                                                     gsparams=gsparams,
                                                     propagate_gsparams=propagate_gsparams)

        else:
            # obj is a SimpleChromaticTransformation
            # Try to keep it that way if possible.
            if not hasattr(flux_ratio,'__call__'):
                # Then transformations are all achromatic.  Apply to original.
                new_obj = Transformation(obj.original, jac, offset, flux_ratio, gsparams,
                                         propagate_gsparams)
                return chrom.SimpleChromaticTransformation(new_obj, obj._flux_ratio,
                                                           gsparams=gsparams,
                                                           propagate_gsparams=propagate_gsparams)

            elif jac is None and offset == (0.,0.):
                # Just a flux_ratio.  Apply to SED.
                new_sed = obj._flux_ratio * flux_ratio
                return chrom.SimpleChromaticTransformation(obj.original, new_sed,
                                                           gsparams=gsparams,
                                                           propagate_gsparams=propagate_gsparams)

            else:
                # Too complicated for simple.
                return chrom.ChromaticTransformation(obj, jac, offset, flux_ratio,
                                                     gsparams=gsparams,
                                                     propagate_gsparams=propagate_gsparams)

    elif not isinstance(obj, (GSObject, chrom.ChromaticObject)):
        raise TypeError("Argument to Transform must be either a GSObject or a ChromaticObject.")

    else:
        # Now we know we are dealing with a more complicated chromatic transformation.

        # Sometimes for Chromatic compound types, it is more efficient to apply the
        # transformation to the components rather than the whole.  In particular, this can
        # help preserve separability in many cases.

        # Don't transform ChromaticSum object, better to just transform the arguments.
        if isinstance(obj, chrom.ChromaticSum):
            new_obj = chrom.ChromaticSum(
                [ Transform(o,jac,offset,flux_ratio,gsparams,propagate_gsparams)
                  for o in obj.obj_list ])
            if hasattr(obj, 'covspec'):
                if jac is None:
                    new_obj.covspec = obj.covspec * flux_ratio**2
                else:
                    dudx, dudy, dvdx, dvdy = np.asarray(jac, dtype=float).flatten()
                    new_obj.covspec = obj.covspec.transform(dudx, dudy, dvdx, dvdy) * flux_ratio**2
            return new_obj

        # If we are just flux scaling, then a Convolution can do that to the first element.
        # NB. Even better, if the flux scaling is chromatic, would be to find a component
        # that is already non-separable.  But we don't bother trying to do that currently.
        elif (isinstance(obj, chrom.ChromaticConvolution or isinstance(obj, convolve.Convolution))
              and jac is None and offset == (0.,0.)):
            first = Transform(obj.obj_list[0], flux_ratio=flux_ratio, gsparams=gsparams,
                              propagate_gsparams=propagate_gsparams)
            return chrom.ChromaticConvolution( [first] + [o for o in obj.obj_list[1:]] )

        else:
            return chrom.ChromaticTransformation(obj, jac, offset, flux_ratio, gsparams=gsparams,
                                                 propagate_gsparams=propagate_gsparams)


class Transformation(GSObject):
    """A class for modeling an affine transformation of a `GSObject` instance.

    Typically, you do not need to construct a Transformation object explicitly.  This is the type
    returned by the various transformation methods of `GSObject` such as `GSObject.shear`,
    `GSObject.rotate`, `GSObject.shift`, `GSObject.transform`, etc.

    All the various transformations can be described as a combination of a jacobian matrix
    (i.e. `GSObject.transform`) and a translation (`GSObject.shift`), which are described by
    (dudx,dudy,dvdx,dvdy) and (dx,dy) respectively.

    .. note::
        The name of the flux_ratio parameter is technically wrong here if the jacobian has a
        non-unit determinant, since that would also scale the flux.  The flux_ratio parameter
        actually only refers to an overall amplitude ratio for the surface brightness profile.
        The total flux scaling is actually ``|det(jac)| * flux_ratio``.

    Parameters:
        obj:                The object to be transformed.
        jac:                A list, tuple or numpy array ( dudx, dudy, dvdx, dvdy ) describing
                            the Jacobian of the transformation.  Use None to indicate that the
                            Jacobian is the 2x2 unit matrix.  [default: None]
        offset:             A galsim.PositionD or tuple giving the offset by which to shift the
                            profile. [default: (0.,0.)]
        flux_ratio:         A factor by which to multiply the surface brightness of the object.
                            (Technically, not necessarily the flux.  See above.) [default: 1]
        gsparams:           An optional `GSParams` argument. [default: None]
        propagate_gsparams: Whether to propagate gsparams to the transformed object.  This
                            is normally a good idea, but there may be use cases where one
                            would not want to do this. [default: True]

    Attributes:
        original:       The original object that is being transformed.
        jac:            The jacobian of the transformation matrix.
        offset:         The offset being applied.
        flux_ratio:     The amount by which the overall surface brightness amplitude is multiplied.
        gsparams:       The usual gsparams attribute that all `GSObject` classes have.

    Note: if ``gsparams`` is unspecified (or None), then the Transformation instance inherits the
    `GSParams` from obj.  Also, note that parameters related to the Fourier-space calculations must
    be set when initializing obj, NOT when creating the Transform (at which point the accuracy and
    threshold parameters will simply be ignored).
    """
    unit_jac = np.array([[1,0],[0,1]], dtype=float)

    def __init__(self, obj, jac=None, offset=(0.,0.), flux_ratio=1.,
                 gsparams=None, propagate_gsparams=True):
        if jac is None:
            self._jac = None
        else:
            self._jac = np.asarray(jac, dtype=float).reshape(2,2)
        offset = PositionD(offset)
        self._dx = offset.x
        self._dy = offset.y
        self._flux_ratio = float(flux_ratio)
        self._gsparams = GSParams.check(gsparams, obj.gsparams)
        self._propagate_gsparams = propagate_gsparams
        if self._propagate_gsparams:
            obj = obj.withGSParams(self._gsparams)

        if isinstance(obj, Transformation):
            # Combine the two affine transformations into one.
            if obj._has_offset:
                if jac is None:
                    dx1, dy1 = obj._dx, obj._dy
                else:
                    dx1, dy1 = self._fwd_normal(obj._dx, obj._dy)
                self._dx += dx1
                self._dy += dy1
            if jac is None:
                self._jac = obj._jac
            else:
                self._jac = self._jac if obj._jac is None else self._jac.dot(obj.jac)
            self._flux_ratio *= obj._flux_ratio
            self._original = obj._original
        else:
            self._original = obj
        self._has_offset = (self._dx != 0. or self._dy != 0.)
        if self._det==0.:
            raise GalSimError("Attempt to Transform with degenerate matrix");

    @property
    def original(self):
        """The original object being transformed.
        """
        return self._original

    @property
    def jac(self):
        """The Jacobian of the transforamtion.
        """
        return self.unit_jac if self._jac is None else self._jac

    @property
    def offset(self):
        """The offset of the transformation.
        """
        return _PositionD(self._dx, self._dy)

    @property
    def flux_ratio(self):
        """The flux ratio of the transformation.
        """
        return self._flux_ratio

    @lazy_property
    def _flux(self):
        return self._flux_scaling * self._original.flux

    @lazy_property
    def _sbp(self):
        _jac = 0 if self._jac is None else self._jac.__array_interface__['data'][0]
        return _galsim.SBTransform(self._original._sbp, _jac,
                                   self._dx, self._dy, self._flux_ratio, self.gsparams._gsp)

    @lazy_property
    def _noise(self):
        if self.original.noise is None:
            return None
        else:
            return BaseCorrelatedNoise(
                    self.original.noise.rng,
                    _Transform(self.original.noise._profile, self._jac,
                               flux_ratio=self.flux_ratio**2),
                    self.original.noise.wcs)

    def withGSParams(self, gsparams=None, **kwargs):
        """Create a version of the current object with the given gsparams

        .. note::

            Unless you set ``propagate_gsparams=False``, this method will also update the gsparams
            of the wrapped component object.
        """
        if gsparams == self.gsparams: return self
        ret = copy.copy(self)
        ret._gsparams = GSParams.check(gsparams, self.gsparams, **kwargs)
        if self._propagate_gsparams:
            ret._original = self._original.withGSParams(ret._gsparams)
        return ret

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, Transformation) and
                 self.original == other.original and
                 np.array_equal(self.jac, other.jac) and
                 np.array_equal(self.offset, other.offset) and
                 self.flux_ratio == other.flux_ratio and
                 self.gsparams == other.gsparams and
                 self._propagate_gsparams == other._propagate_gsparams))

    def __hash__(self):
        return hash(("galsim.Transformation", self.original, tuple(self.jac.ravel()),
                     self._dx, self._dy, self.flux_ratio, self.gsparams,
                     self._propagate_gsparams))

    def __repr__(self):
        return ('galsim.Transformation(%r, jac=%r, offset=%r, flux_ratio=%r, gsparams=%r, '
                'propagate_gsparams=%r)')%(
            self.original, None if self._jac is None else self._jac.tolist(),
            self.offset, self.flux_ratio, self.gsparams, self._propagate_gsparams)

    @classmethod
    def _str_from_jac(cls, jac):
        dudx, dudy, dvdx, dvdy = jac.ravel()
        # Figure out the shear/rotate/dilate calls that are equivalent.
        jac = JacobianWCS(dudx,dudy,dvdx,dvdy)
        scale, shear, theta, flip = jac.getDecomposition()
        s = None
        if flip:
            s = 0  # Special value indicating to just use transform.
        if abs(theta.rad) > 1.e-12:
            if s is None:
                s = '.rotate(%s)'%theta
            else:
                s = 0
        if shear.g > 1.e-12:
            if s is None:
                s = '.shear(%s)'%shear
            else:
                s = 0
        if abs(scale-1.0) > 1.e-12:
            if s is None:
                s = '.expand(%s)'%scale
            else:
                s = 0
        if s == 0:
            # If flip or there are two components, then revert to transform as simpler.
            s = '.transform(%s,%s,%s,%s)'%(dudx,dudy,dvdx,dvdy)
        if s is None:
            # If nothing is large enough to show up above, give full detail of transform
            s = '.transform(%r,%r,%r,%r)'%(dudx,dudy,dvdx,dvdy)
        return s

    def __str__(self):
        s = str(self.original)
        if self._jac is not None:
            s += self._str_from_jac(self._jac)
        if self._dx != 0 or self._dy != 0:
            s += '.shift(%s,%s)'%(self._dx,self._dy)
        if self.flux_ratio != 1.:
            s += ' * %s'%self.flux_ratio
        return s

    def _prepareDraw(self):
        self._original._prepareDraw()

    # Some lazy properties to calculate things as needed.
    @lazy_property
    def _det(self):
        if self._jac is None:
            return 1
        if self._jac[0,1] == 0. and self._jac[1,0] == 0.:
            if self._jac[0,0] == 1. and self._jac[1,1] == 1.:     # jac is (1,0,0,1)
                return 1.
            else:                                               # jac is (a,0,0,b)
                return self._jac[0,0] * self._jac[1,1]
        else:                                                   # Fully general case
            return self._jac[0,0] * self._jac[1,1] - self._jac[0,1] * self._jac[1,0]

    @lazy_property
    def _invdet(self):
        return 1. if self._jac is None else 1./self._det

    @lazy_property
    def _invjac(self):
        invjac = np.array([[self._jac[1,1], -self._jac[0,1]], [-self._jac[1,0], self._jac[0,0]]])
        invjac *= self._invdet
        return invjac

    # To avoid confusion with the flux vs amplitude scaling, we use these names below, rather
    # than flux_ratio, which is really an amplitude scaling.
    @property
    def _amp_scaling(self):
        return self._flux_ratio

    @lazy_property
    def _flux_scaling(self):
        return abs(self._det) * self._flux_ratio

    # Some helper attributes to make fwd and inv transformation quicker
    @lazy_property
    def _fwd(self):
        if self._jac is None:
            return WeakMethod(self._ident)
        elif self._jac[0,1] == 0. and self._jac[1,0] == 0.:
            return WeakMethod(self._fwd_diag)
        else:
            return WeakMethod(self._fwd_normal)

    @lazy_property
    def _fwdT(self):
        if self._jac is None:
            return WeakMethod(self._ident)
        elif self._jac[0,1] == 0. and self._jac[1,0] == 0.:
            return WeakMethod(self._fwd_diag)
        else:
            return WeakMethod(self._fwdT_normal)

    @lazy_property
    def _inv(self):
        if self._jac is None:
            return WeakMethod(self._ident)
        elif self._jac[0,1] == 0. and self._jac[1,0] == 0.:
            return WeakMethod(self._inv_diag)
        else:
            return WeakMethod(self._inv_normal)

    @lazy_property
    def _kfactor(self):
        if self._dx == self._dy == 0.:
            return WeakMethod(self._kf_nophase)
        else:
            return WeakMethod(self._kf_phase)

    def _ident(self, x, y):
        return x, y

    def _fwd_diag(self, x, y):
        x *= self._jac[0,0]
        y *= self._jac[1,1]
        return x, y

    def _fwd_normal(self, x, y):
        #return self._jac[0,0] * x + self._jac[0,1] * y, self._jac[1,0] * x + self._jac[1,1] * y
        # Do this as much in place as possible
        temp = self._jac[0,1] * y
        y *= self._jac[1,1]
        y += self._jac[1,0] * x
        x *= self._jac[0,0]
        x += temp
        return x, y

    def _fwdT_normal(self, x, y):
        #return self._jac[0,0] * x + self._jac[1,0] * y, self._jac[0,1] * x + self._jac[1,1] * y
        temp = self._jac[1,0] * y
        y *= self._jac[1,1]
        y += self._jac[0,1] * x
        x *= self._jac[0,0]
        x += temp
        return x, y

    def _inv_diag(self, x, y):
        x /= self._jac[0,0]
        y /= self._jac[1,1]
        return x, y

    def _inv_normal(self, x, y):
        #return (self._invdet * (self._jac[1,1] * x - self._jac[0,1] * y),
        #        self._invdet * (-self._jac[1,0] * x + self._jac[0,0] * y))
        temp = self._invjac[0,1] * y
        y *= self._invjac[1,1]
        y += self._invjac[1,0] * x
        x *= self._invjac[0,0]
        x += temp
        return x, y

    def _kf_nophase(self, kx, ky):
        return self._flux_scaling

    def _kf_phase(self, kx, ky):
        kx *= -1j * self._dx
        ky *= -1j * self._dy
        kx += ky
        return self._flux_scaling * np.exp(kx)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        d.pop('_fwd',None)
        d.pop('_fwdT',None)
        d.pop('_inv',None)
        d.pop('_kfactor',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    def _major_minor(self):
        if not hasattr(self, "_major"):
            if self._jac is None:
                self._major = self._minor = 1.
            else:
                h1 = math.hypot(self._jac[0,0] + self._jac[1,1], self._jac[0,1] - self._jac[1,0])
                h2 = math.hypot(self._jac[0,0] - self._jac[1,1], self._jac[0,1] + self._jac[1,0])
                self._major = 0.5 * abs(h1+h2)
                self._minor = 0.5 * abs(h1-h2)

    @lazy_property
    def _maxk(self):
        self._major_minor()
        return self._original.maxk / self._minor

    @lazy_property
    def _stepk(self):
        self._major_minor()
        stepk = self._original.stepk / self._major
        # If we have a shift, we need to further modify stepk
        #     stepk = Pi/R
        #     R <- R + |shift|
        #     stepk <- Pi/(Pi/stepk + |shift|)
        if self._dx !=  0. or self._dy != 0.:
            dr = math.hypot(self._dx, self._dy)
            stepk = math.pi / (math.pi/stepk + dr)
        return stepk

    @property
    def _has_hard_edges(self):
        return self._original.has_hard_edges

    @property
    def _is_axisymmetric(self):
        return bool(self._original.is_axisymmetric and
                    (self._jac is None or (self._jac[0,0] == self._jac[1,1] and
                                           self._jac[0,1] == -self._jac[1,0])) and
                    self._dx == self._dy == 0.)

    @property
    def _is_analytic_x(self):
        return self._original.is_analytic_x

    @property
    def _is_analytic_k(self):
        return self._original.is_analytic_k

    @property
    def _centroid(self):
        cen = self._original.centroid
        cenx, ceny = self._fwd(cen.x, cen.y)
        cenx += self._dx
        ceny += self._dy
        return _PositionD(cenx,ceny)

    @property
    def _positive_flux(self):
        return self._flux_scaling * self._original.positive_flux

    @property
    def _negative_flux(self):
        return self._flux_scaling * self._original.negative_flux

    @lazy_property
    def _flux_per_photon(self):
        return self._calculate_flux_per_photon()

    @property
    def _max_sb(self):
        return self._amp_scaling * self._original.max_sb

    def _xValue(self, pos):
        x = pos.x - self._dx
        y = pos.y - self._dy
        inv_pos = _PositionD(*self._inv(x, y))
        return self._original._xValue(inv_pos) * self._amp_scaling

    def _kValue(self, kpos):
        fwdT_kpos = _PositionD(*self._fwdT(kpos.x, kpos.y))
        return self._original._kValue(fwdT_kpos) * self._kfactor(kpos.x, kpos.y)

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        dx, dy = offset
        if self._has_offset:
            if jac is not None:
                x1 = jac[0,0] * self._dx + jac[0,1] * self._dy
                y1 = jac[1,0] * self._dx + jac[1,1] * self._dy
            else:
                x1, y1 = self._dx, self._dy
            dx += x1
            dy += y1
        flux_scaling *= self._flux_scaling
        jac = self._jac if jac is None else jac if self._jac is None else jac.dot(self._jac)
        self._original._drawReal(image, jac, (dx, dy), flux_scaling)

    def _shoot(self, photons, rng):
        self._original._shoot(photons, rng)
        photons.x, photons.y = self._fwd(photons.x, photons.y)
        photons.x += self._dx
        photons.y += self._dy
        photons.scaleFlux(self._flux_scaling)

    def _drawKImage(self, image, jac=None):
        jac1 = self._jac if jac is None else jac if self._jac is None else jac.dot(self._jac)
        self._original._drawKImage(image, jac1)

        if self._has_offset:
            _jac = 0 if jac is None else jac.__array_interface__['data'][0]
            _galsim.ApplyKImagePhases(image._image, image.scale, _jac,
                                      self._dx, self._dy, self._flux_scaling)
        elif abs(self._flux_scaling-1.) > self._gsparams.kvalue_accuracy:
            image *= self._flux_scaling


def _Transform(obj, jac=None, offset=(0.,0.), flux_ratio=1.):
    """Approximately equivalent to Transform, but without some of the sanity checks (such as
    checking for chromatic options) or setting a new gsparams.

    For a `ChromaticObject`, you must use the regular `Transform`.

    Parameters:
        obj:                The object to be transformed.
        jac:                A 2x2 numpy array describing the Jacobian of the transformation.
                            Use None to indicate that the Jacobian is the 2x2 unit matrix.
                            [default: None]
        offset:             The offset to apply. [default (0.,0.)]
        flux_ratio:         A factor by which to multiply the surface brightness of the object.
                            [default: 1.]
    """
    ret = Transformation.__new__(Transformation)
    ret._gsparams = obj.gsparams
    ret._propagate_gsparams = True
    ret._jac = jac
    ret._dx, ret._dy = offset
    if isinstance(obj, Transformation):
        if obj._has_offset:
            if jac is None:
                dx1, dy1 = obj._dx, obj._dy
            else:
                dx1, dy1 = ret._fwd_normal(obj._dx, obj._dy)
            ret._dx += dx1
            ret._dy += dy1
        if jac is None:
            ret._jac = obj._jac
        else:
            ret._jac = ret._jac if obj._jac is None else ret._jac.dot(obj.jac)
        ret._flux_ratio = flux_ratio * obj._flux_ratio
        ret._original = obj._original
    else:
        ret._flux_ratio = flux_ratio
        ret._original = obj
    ret._has_offset = (ret._dx != 0. or ret._dy != 0.)
    return ret

# Put this at the bottom to avoid circular import error.
from .correlatednoise import BaseCorrelatedNoise



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#
"""
A program to download the COSMOS RealGalaxy catalog for use with GalSim.
"""

import os, sys, tarfile, subprocess, shutil, json
import argparse
import logging
from urllib.request import urlopen

from ._version import __version__ as version
from .meta_data import share_dir
from .utilities import ensure_dir
from .main import make_logger

script_name = 'galsim_download_cosmos'

def parse_args(command_args):
    """Handle the command line arguments using either argparse (if available) or optparse.
    """
    # Another potential option we might want to add is to download the smaller training sample
    # rather than the full 4 GB file.  Right now, this just downloads the larger catalog.

    # Short description strings common to both parsing mechanisms
    description = "This program will download the COSMOS RealGalaxy catalog and images\n"
    description += "and place them in the GalSim share directory so they can be used as\n "
    description += "the default files for the RealGalaxyCatalog class.\n"
    description += "See https://github.com/GalSim-developers/GalSim/wiki/RealGalaxy%20Data\n"
    description += "for more details about the files being downloaded."
    epilog = "Note: The unpacked files total almost 6 GB in size!\n"

    # Build the parser and add arguments
    parser = argparse.ArgumentParser(description=description, epilog=epilog, add_help=True)
    parser.add_argument(
        '-v', '--verbosity', type=int, action='store', default=2, choices=(0, 1, 2, 3),
        help='Integer verbosity level: min=0, max=3 [default=2]')
    parser.add_argument(
        '-f', '--force', action='store_const', default=False, const=True,
        help='Force overwriting the current file if one exists')
    parser.add_argument(
        '-q', '--quiet', action='store_const', default=False, const=True,
        help="Don't ask about re-downloading an existing file. (implied by verbosity=0)")
    parser.add_argument(
        '-u', '--unpack', action='store_const', default=False, const=True,
        help='Re-unpack the tar file if not downloading')
    parser.add_argument(
        '--save', action='store_const', default=False, const=True,
        help="Save the tarball after unpacking.")
    parser.add_argument(
        '-d', '--dir', action='store', default=None,
        help="Install into an alternate directory and link from the share/galsim directory")
    parser.add_argument(
        '-s', '--sample', action='store', default='25.2', choices=('23.5', '25.2'),
        help="Flux limit for sample to download; either 23.5 or 25.2")
    parser.add_argument(
        '--nolink', action='store_const', default=False, const=True,
        help="Don't link to the alternate directory from share/galsim")
    args = parser.parse_args(command_args)
    args.log_file = None
    args.log_format = None

    # Return the args
    return args

def get_input():  # pragma: no cover
    # A level of indirection to make it easier to test functions using input.
    # This one isn't covered, since we always mock it.
    return input()

# Based on recipe 577058: http://code.activestate.com/recipes/577058/
def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is one of "yes" or "no".
    """
    valid = {"yes":"yes",   "y":"yes",  "ye":"yes",
             "no":"no",     "n":"no"}
    if default == None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while 1:
        sys.stdout.write(question + prompt)
        choice = get_input().lower()
        if default is not None and choice == '':
            choice = default
            break
        elif choice in valid.keys():
            break
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "\
                             "(or 'y' or 'n').\n")
    return valid[choice]

def get_names(args, logger):
    if args.dir is not None:
        target_dir = os.path.expanduser(args.dir)
        do_link = not args.nolink
    else:
        target_dir = share_dir
        do_link = False

    url = "https://zenodo.org/record/3242143/files/COSMOS_%s_training_sample.tar.gz"%(
            args.sample)
    file_name = os.path.basename(url)
    target = os.path.join(target_dir, file_name)
    link_dir = os.path.join(share_dir, file_name)[:-len('.tar.gz')]
    unpack_dir = target[:-len('.tar.gz')]
    logger.warning('Downloading from url:\n  %s',url)
    logger.warning('Target location is:\n  %s',target)

    return url, target, target_dir, link_dir, unpack_dir, do_link

def get_meta(url, args, logger):
    logger.info('')

    # See how large the file to be downloaded is.
    u = urlopen(url)
    meta = u.info()
    logger.debug("Meta information about url:\n%s",str(meta))
    file_name = os.path.basename(url)
    file_size = int(meta.get("Content-Length"))
    logger.info("Size of %s: %d MBytes" , file_name, file_size/1024**2)

    return meta

def check_existing(target, unpack_dir, meta, args, logger):
    # Make sure the directory we want to put this file exists.
    ensure_dir(target)

    do_download = True
    # If file already exists
    if os.path.isfile(target):
        file_size = int(meta.get("Content-Length"))
        logger.info("")
        existing_file_size = os.path.getsize(target)
        if args.force:
            logger.info("Target file already exists.  Size = %d MBytes.  Forced re-download.",
                        existing_file_size/1024**2)
        elif file_size == existing_file_size:
            if args.quiet:
                logger.info("Target file already exists.  Not re-downloading.")
                do_download = False
            else:
                q = "Target file already exists.  Overwrite?"
                yn = query_yes_no(q, default='no')
                if yn == 'no':
                    do_download = False
        else:
            logger.warning("Target file already exists, but it seems to be either incomplete, "
                           "corrupt, or obsolete")
            if args.quiet:
                logger.info("Size of existing file = %d MBytes.  Re-downloading.",
                            existing_file_size/1024**2)
            else:
                q = "Size of existing file = %d MBytes.  Re-download?"%(existing_file_size/1024**2)
                yn = query_yes_no(q, default='yes')
                if yn == 'no':
                    do_download = False
    elif unpack_dir is not None and os.path.isdir(unpack_dir):
        logger.info("")

        # Check that this is the current version.
        meta_file = os.path.join(unpack_dir, 'meta.json')
        logger.debug('meta_file = %s',meta_file)
        if os.path.isfile(meta_file):
            logger.debug('meta_file exists')
            with open(meta_file) as fp:
                saved_meta_dict = json.load(fp)
                # Get rid of the unicode
                saved_meta_dict = dict([ (str(k),str(v)) for k,v in saved_meta_dict.items()])
            saved_meta_dict = {k.lower():v for k,v in saved_meta_dict.items()}
            logger.debug("current meta information is %s",saved_meta_dict)
            meta_dict = dict(meta)
            meta_dict = {k.lower():v for k,v in meta_dict.items()}
            logger.debug("url's meta information is %s",meta_dict)

            # There are several checksum tags.  If any of them are present and match,
            # then file is current.  If they don't match, file is obsolete.
            # If none are present, then they changed something, so do a longer check.
            # (And try to get around to updating this list of checksum keys.)
            checksum_keys = ['oc-checksum', 'content-md5', 'etag']
            obsolete = None
            for k in checksum_keys:
                if k in saved_meta_dict and k in meta_dict:
                    if saved_meta_dict[k] == meta_dict[k]:
                        logger.info("Checksum key %s matches.  File is up to date."%k)
                        obsolete = False
                    else:
                        logger.info("Checksum key %s doesn't matches.  File is obsolete."%k)
                        obsolete = True
                    break

            if obsolete is None:
                # Check all meta data.  If it all matches, then it's fine.
                # Otherwise, the file is obsolete.
                obsolete = False
                for k in meta_dict:
                    # Skip some keys that don't imply obselescence.
                    if k.startswith('x-') or k.startswith('retry') or k.startswith('set-cookie'):
                        continue
                    if k == 'date' or k == 'last-modified' or k == 'server':
                        continue
                    # Others that are missing or different imply obsolete
                    if k not in saved_meta_dict:
                        logger.debug("key %s is missing in saved meta information",k)
                        obsolete = True
                    elif meta_dict[k] != saved_meta_dict[k]:
                        logger.debug("key %s differs: %s != %s",k,meta_dict[k],saved_meta_dict[k])
                        obsolete = True
                    else:
                        logger.debug("key %s matches",k)
        else:
            logger.debug('meta_file does not exist')
            obsolete = True

        if obsolete:
            if args.quiet or args.force:
                logger.warning("The version currently on disk is obsolete.  "
                               "Downloading new version.")
            else:
                q = "The version currently on disk is obsolete.  Download new version?"
                yn = query_yes_no(q, default='yes')
                if yn == 'no':
                    do_download = False
        elif args.force:
            logger.info("Target file has already been downloaded and unpacked.  "
                        "Forced re-download.")
        elif args.quiet:
            logger.info("Target file has already been downloaded and unpacked.  "
                        "Not re-downloading.")
            args.save = True  # Don't delete it!
            do_download = False
        else:
            q = "Target file has already been downloaded and unpacked.  Re-download?"
            yn = query_yes_no(q, default='no')
            if yn == 'no':
                args.save = True
                do_download = False
    return do_download

def download(do_download, url, target, meta, args, logger):
    if not do_download: return
    logger.info("")
    u = urlopen(url)
    # This bit is based on one of the answers here: (by PabloG)
    # http://stackoverflow.com/questions/22676/how-do-i-download-a-file-over-http-using-python
    # The progress feature in that answer is important here, since downloading such a large file
    # will take a while.
    file_size = int(meta.get("Content-Length"))
    try:
        with open(target, 'wb') as f:
            file_size_dl = 0
            block_sz = 32 * 1024
            next_dot = file_size/100.  # For verbosity==1, the next size for writing a dot.
            while True:
                buffer = u.read(block_sz)
                if not buffer:
                    break

                file_size_dl += len(buffer)
                f.write(buffer)

                # Status bar
                if args.verbosity >= 2:
                    status = r"Downloading: %5d / %d MBytes  [%3.2f%%]" % (
                        file_size_dl/1024**2, file_size/1024**2, file_size_dl*100./file_size)
                    status = status + '\b'*len(status)
                    sys.stdout.write(status)
                    sys.stdout.flush()
                elif args.verbosity >= 1 and file_size_dl > next_dot:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                    next_dot += file_size/100.
        logger.info("Download complete.")
        logger.info("")
    except OSError as e:
        # Try to give a reasonable suggestion for some common OSErrors.
        logger.error("\n\nOSError: %s",str(e))
        if 'Permission denied' in str(e):
            logger.error("Rerun using sudo %s",script_name)
            logger.error("If this is not possible, you can download to an alternate location:")
            logger.error("    %s -d dir_name --nolink\n",script_name)
        elif 'Disk quota' in str(e) or 'No space' in str(e):
            logger.error("You might need to download this in an alternate location and link:")
            logger.error("    %s -d dir_name\n",script_name)
        raise

def check_unpack(do_download, unpack_dir, target, args):
    # Usually we unpack if we downloaded the tarball or if specified by the command line option.
    do_unpack = do_download or args.unpack

    # If the unpack dir is missing, then need to unpack
    if not os.path.exists(unpack_dir):
        do_unpack = True

    # But of course if there is no tarball, we can't unpack it
    if not os.path.isfile(target):
        do_unpack = False

    # If we have a downloaded tar file, ask if it should be re-unpacked.
    if not do_unpack and not args.quiet and os.path.isfile(target):
        q = "Tar file is already unpacked.  Re-unpack?"
        yn = query_yes_no(q, default='no')
        if yn == 'yes':
            do_unpack=True
    return do_unpack

def unpack(do_unpack, target, target_dir, unpack_dir, meta, args, logger):
    if not do_unpack: return
    logger.info("Unpacking the tarball...")
    with tarfile.open(target) as tar:
        if args.verbosity >= 3:
            tar.list(verbose=True)
        elif args.verbosity >= 2:
            tar.list(verbose=False)

        def is_within_directory(directory, target):
            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)
            prefix = os.path.commonprefix([abs_directory, abs_target])
            return prefix == abs_directory

        def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
            # This is somewhat gratuitous for our use case, since we directly control the tar file,
            # but apparently there is a security vulnerability in the tar.extractall function.
            # This bit of code was added by the Advanced Research Center at Trellix in PR #1188.
            # For more information about the security vulnerability, see
            # https://github.com/advisories/GHSA-gw9q-c7gh-j9vm

            for member in tar.getmembers():
                member_path = os.path.join(path, member.name)
                if not is_within_directory(path, member_path):  # pragma: no cover
                    raise Exception("Attempted Path Traversal in Tar File")

            if hasattr(tarfile, 'data_filter'):
                # New feature in py3.12.  If you don't use it, you get a deprecation warning.
                tar.extractall(path, members, numeric_owner=numeric_owner, filter='data')
            else:
                tar.extractall(path, members, numeric_owner=numeric_owner)

        safe_extract(tar, target_dir)

    # Write the meta information to a file, meta.json to mark what version this all is.
    meta_file = os.path.join(unpack_dir, 'meta.json')
    with open(meta_file,'w') as fp:
        json.dump(dict(meta), fp)

    logger.info("Extracted contents of tar file.")
    logger.info("")

def check_remove(do_unpack, target, args):
    # Usually, we remove the tarball if we unpacked it and command line doesn't specify to save it.
    do_remove = do_unpack and not args.save

    # But if we didn't unpack it, and they didn't say to save it, ask if we should remove it.
    if os.path.isfile(target) and not do_remove and not args.save and not args.quiet:
        q = "Remove the tarball?"
        yn = query_yes_no(q, default='no')
        if yn == 'yes':
            do_remove = True
    return do_remove

def remove_tarball(do_remove, target, logger):
    if do_remove:
        logger.info("Removing the tarball to save space")
        os.remove(target)

def make_link(do_link, unpack_dir, link_dir, args, logger):
    if not do_link: return
    logger.debug("Linking to %s from %s", unpack_dir, link_dir)
    if os.path.lexists(link_dir):
        if os.path.islink(link_dir):
            # If it exists and is a link, we just remove it and relink without any fanfare.
            logger.debug("Removing existing link")
            os.unlink(link_dir)
        else:
            # If it is not a link, we need to figure out what to do with it.
            if os.path.isdir(link_dir):
                # If it's a directory, probably want to keep it.
                logger.warning("%s already exists and is a directory.",link_dir)
                if args.force:
                    logger.warning("Removing the existing files to make the link.")
                elif args.quiet:
                    logger.warning("Link cannot be made.  (Use -f to force removal of existing dir.)")
                    return
                else:
                    q = "Remove the existing files to make the link?"
                    yn = query_yes_no(q, default='no')
                    if yn == 'no':
                        return
                shutil.rmtree(link_dir)
            else:
                # If it's not a directory, it's probably corrupt, so the default is to remove it.
                logger.warning("%s already exists, but strangely isn't a directory.",link_dir)
                if args.force or args.quiet:
                    logger.warning("Removing the existing file.")
                else:
                    q = "Remove the existing file?"
                    yn = query_yes_no(q, default='yes')
                    if yn == 'no':
                        return
                os.remove(link_dir)
    os.symlink(os.path.abspath(unpack_dir), link_dir)
    logger.info("Made link to %s from %s", unpack_dir, link_dir)


def download_cosmos(args, logger):
    """The main script given the ArgParsed args and a logger
    """
    # Give diagnostic about GalSim version
    logger.debug("GalSim version: %s",version)
    logger.debug("This download script is: %s",__file__)
    logger.info("Type %s -h to see command line options.\n",script_name)

    # Some definitions:
    # share_dir is the base galsim share directory, e.g. /usr/local/share/galsim/
    # url is the url from which we will download the tarball.
    # target is the full path of the downloaded tarball
    # target_dir is where we will put the downloaded file, usually == share_dir.
    # link_dir is the directory where this would normally have been unpacked.
    # unpack_dir is the directory that the tarball will unpack into.

    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)

    meta = get_meta(url, args, logger)

    # Check if the file already exists and if it is the right size
    do_download = check_existing(target, unpack_dir, meta, args, logger)

    # Download the tarball
    download(do_download, url, target, meta, args, logger)

    # Unpack the tarball
    do_unpack = check_unpack(do_download, unpack_dir, target, args)
    unpack(do_unpack, target, target_dir, unpack_dir, meta, args, logger)

    # Remove the tarball
    do_remove = check_remove(do_unpack, target, args)
    remove_tarball(do_remove, target, logger)

    # If we are downloading to an alternate directory, we (usually) link to it from share/galsim
    make_link(do_link, unpack_dir, link_dir, args, logger)

def main(command_args):
    """The whole process given command-line parameters in their native (non-ArgParse) form.
    """
    args = parse_args(command_args)
    logger = make_logger(args)
    download_cosmos(args, logger)

def run_main():
    """Kick off the process grabbing the command-line parameters from sys.argv
    """
    main(sys.argv[1:])



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Add', 'Sum', ]

import numpy as np
import copy

from .gsparams import GSParams
from .gsobject import GSObject
from .utilities import lazy_property
from . import _galsim
from .photon_array import PhotonArray
from .random import UniformDeviate
from . import chromatic as chrom


def Add(*args, **kwargs):
    """A function for adding 2 or more `GSObject` or `ChromaticObject` instances.

    This function will inspect its input arguments to decide if a `Sum` object or a
    `ChromaticSum` object is required to represent the sum of surface brightness profiles.

    Typically, you do not need to call Add() explicitly.  Normally, you would just use the +
    operator, which returns a Sum::

        >>> bulge = galsim.Sersic(n=3, half_light_radius=0.8)
        >>> disk = galsim.Exponential(half_light_radius=1.4)
        >>> gal = bulge + disk
        >>> psf = galsim.Gaussian(sigma=0.3, flux=0.3) + galsim.Gaussian(sigma=0.8, flux=0.7)

    If one of the items is chromatic, it will return a `ChromaticSum`::

        >>> disk = galsim.Exponential(half_light_radius=1.4) * galsim.SED(sed_file)
        >>> gal = bulge + disk

    Parameters:
        args:               Unnamed args should be a list of objects to add.
        gsparams:           An optional `GSParams` argument. [default: None]
        propagate_gsparams: Whether to propagate gsparams to each of the components.  This
                            is normally a good idea, but there may be use cases where one
                            would not want to do this. [default: True]

    Returns:
        a `Sum` or `ChromaticSum` instance as appropriate.
    """
    if len(args) == 0:
        raise TypeError("At least one ChromaticObject or GSObject must be provided.")
    elif len(args) == 1:
        # 1 argument.  Should be either a GSObject or a list of GSObjects
        if isinstance(args[0], (GSObject, chrom.ChromaticObject)):
            args = [args[0]]
        elif isinstance(args[0], list) or isinstance(args[0], tuple):
            args = args[0]
        else:
            raise TypeError("Single input argument must be a GSObject, ChromaticObject or "
                            + "a (possibly mixed) list of them.")
    # else args is already the list of objects

    if any([isinstance(a, chrom.ChromaticObject) for a in args]):
        return chrom.ChromaticSum(*args, **kwargs)
    else:
        return Sum(*args, **kwargs)


class Sum(GSObject):
    """A class for adding 2 or more `GSObject` instances.

    The Sum class is used to represent the sum of multiple `GSObject` instances.  For example, it
    might be used to represent a multiple-component galaxy as the sum of an Exponential and a
    DeVaucouleurs, or to represent a PSF as the sum of multiple Gaussian objects.

    Typically, you do not need to construct a Sum object explicitly.  Normally, you would just
    use the + operator, which returns a Sum::

        >>> bulge = galsim.Sersic(n=3, half_light_radius=0.8)
        >>> disk = galsim.Exponential(half_light_radius=1.4)
        >>> gal = bulge + disk
        >>> psf = galsim.Gaussian(sigma=0.3, flux=0.3) + galsim.Gaussian(sigma=0.8, flux=0.7)

    You can also use the Add() factory function, which returns a Sum object if none of the
    individual objects are chromatic::

        >>> gal = galsim.Add([bulge,disk])

    Parameters:
        args:               Unnamed args should be a list of objects to add.
        gsparams:           An optional `GSParams` argument. [default: None]
        propagate_gsparams: Whether to propagate gsparams to each of the components.  This
                            is normally a good idea, but there may be use cases where one
                            would not want to do this. [default: True]

    Note: if ``gsparams`` is unspecified (or None), then the Sum instance will use the most
    restrictive combination of parameters from each of the component objects. Normally, this means
    the smallest numerical value (e.g. folding_threshold, xvalue_accuracy, etc.), but for a few
    parameters, the largest numerical value is used.  See `GSParams.combine` for details.

    Furthermore, the gsparams used for the Sum (either given explicitly or derived from the
    components) will normally be applied to each of the components.  It doesn't usually make much
    sense to apply stricter-than-normal accuracy or threshold values to one component but not
    another in a Sum, so this ensures that they all have consistent rendering behavior.
    However, if you want to keep the existing gsparams of the component objects (e.g. because
    one object is much fainter and can thus use looser accuracy tolerances), then you may
    set ``propagate_gsparams=False``.
    """
    def __init__(self, *args, **kwargs):

        # Check kwargs first:
        gsparams = kwargs.pop("gsparams", None)
        self._propagate_gsparams = kwargs.pop('propagate_gsparams', True)

        # Make sure there is nothing left in the dict.
        if kwargs:
            raise TypeError(
                "Sum constructor got unexpected keyword argument(s): %s"%kwargs.keys())

        if len(args) == 0:
            raise TypeError("At least one ChromaticObject or GSObject must be provided.")
        elif len(args) == 1:
            # 1 argument.  Should be either a GSObject or a list of GSObjects
            if isinstance(args[0], GSObject):
                args = [args[0]]
            elif isinstance(args[0], list) or isinstance(args[0], tuple):
                args = args[0]
            else:
                raise TypeError("Single input argument must be a GSObject or list of them.")
        # else args is already the list of objects

        # Consolidate args for Sums of Sums...
        new_args = []
        for a in args:
            if isinstance(a, Sum):
                new_args.extend(a._obj_list)
            else:
                new_args.append(a)
        args = new_args

        # Save the list as an attribute, so it can be inspected later if necessary.
        self._obj_list = args

        for obj in args:
            if not isinstance(obj, GSObject):
                raise TypeError("Arguments to Sum must be GSObjects, not %s"%obj)

        # Figure out what gsparams to use
        if gsparams is None:
            # If none is given, take the most restrictive combination from the obj_list.
            self._gsparams = GSParams.combine([obj.gsparams for obj in args])
        else:
            # If something explicitly given, then use that.
            self._gsparams = GSParams.check(gsparams)

        # Apply gsparams to all in obj_list.
        if self._propagate_gsparams:
            self._obj_list = [obj.withGSParams(self._gsparams) for obj in args]
        else:
            self._obj_list = args


    @property
    def obj_list(self):
        """The list of objects being added.
        """
        return self._obj_list

    @lazy_property
    def _sbp(self):
        # NB. I only need this until compound and transform are reimplemented in Python...
        sb_list = [obj._sbp for obj in self.obj_list]
        return _galsim.SBAdd(sb_list, self.gsparams._gsp)

    @lazy_property
    def _flux(self):
        flux_list = [obj.flux for obj in self.obj_list]
        return np.sum(flux_list)

    @lazy_property
    def _noise(self):
        # If any of the objects have a noise attribute, then we propagate the sum of the
        # noises (they add like variances) to the final sum.
        _noise = None
        for obj in self.obj_list:
            if obj.noise is not None:
                if _noise is None:
                    _noise = obj.noise
                else:
                    _noise += obj.noise
        return _noise

    def withGSParams(self, gsparams=None, **kwargs):
        """Create a version of the current object with the given gsparams

        .. note::

            Unless you set ``propagate_gsparams=False``, this method will also update the gsparams
            of each object being added.
        """
        if gsparams == self.gsparams: return self
        ret = copy.copy(self)
        ret._gsparams = GSParams.check(gsparams, self.gsparams, **kwargs)
        if self._propagate_gsparams:
            ret._obj_list = [ obj.withGSParams(ret._gsparams) for obj in self.obj_list ]
        return ret

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, Sum) and
                 self.obj_list == other.obj_list and
                 self.gsparams == other.gsparams and
                 self._propagate_gsparams == other._propagate_gsparams))

    def __hash__(self):
        return hash(("galsim.Sum", tuple(self.obj_list), self.gsparams, self._propagate_gsparams))

    def __repr__(self):
        return 'galsim.Sum(%r, gsparams=%r, propagate_gsparams=%r)'%(self.obj_list, self.gsparams,
                self._propagate_gsparams)

    def __str__(self):
        str_list = [ str(obj) for obj in self.obj_list ]
        return '(' + ' + '.join(str_list) + ')'

    def _prepareDraw(self):
        for obj in self.obj_list:
            obj._prepareDraw()

    @lazy_property
    def _maxk(self):
        maxk_list = [obj.maxk for obj in self.obj_list]
        return np.max(maxk_list)

    @lazy_property
    def _stepk(self):
        stepk_list = [obj.stepk for obj in self.obj_list]
        return np.min(stepk_list)

    @lazy_property
    def _has_hard_edges(self):
        hard_list = [obj.has_hard_edges for obj in self.obj_list]
        return bool(np.any(hard_list))

    @lazy_property
    def _is_axisymmetric(self):
        axi_list = [obj.is_axisymmetric for obj in self.obj_list]
        return bool(np.all(axi_list))

    @lazy_property
    def _is_analytic_x(self):
        ax_list = [obj.is_analytic_x for obj in self.obj_list]
        return bool(np.all(ax_list))

    @lazy_property
    def _is_analytic_k(self):
        ak_list = [obj.is_analytic_k for obj in self.obj_list]
        return bool(np.all(ak_list))

    @lazy_property
    def _centroid(self):
        cen_list = [obj.centroid * obj.flux for obj in self.obj_list]
        return sum(cen_list[1:], cen_list[0]) / self.flux

    @lazy_property
    def _positive_flux(self):
        pflux_list = [obj.positive_flux for obj in self.obj_list]
        return np.sum(pflux_list)

    @lazy_property
    def _negative_flux(self):
        nflux_list = [obj.negative_flux for obj in self.obj_list]
        return np.sum(nflux_list)

    @lazy_property
    def _flux_per_photon(self):
        return self._calculate_flux_per_photon()

    @lazy_property
    def _max_sb(self):
        sb_list = [obj.max_sb for obj in self.obj_list]
        return np.sum(sb_list)

    def _xValue(self, pos):
        xv_list = [obj.xValue(pos) for obj in self.obj_list]
        return np.sum(xv_list)

    def _kValue(self, pos):
        kv_list = [obj.kValue(pos) for obj in self.obj_list]
        return np.sum(kv_list)

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        self.obj_list[0]._drawReal(image, jac, offset, flux_scaling)
        if len(self.obj_list) > 1:
            im1 = image.copy()
            for obj in self.obj_list[1:]:
                obj._drawReal(im1, jac, offset, flux_scaling)
                image += im1

    def _shoot(self, photons, rng):
        # We probabilistically choose a component for each photon based on the
        # relative flux density of that component for the given wavelength.
        comp_flux = np.array([obj.positive_flux + obj.negative_flux for obj in self._obj_list])
        total_flux = np.sum(comp_flux)

        prob = comp_flux / total_flux
        cdf = np.cumsum(prob)

        u = np.empty(len(photons))
        UniformDeviate(rng).generate(u)
        use_k = np.sum((u >= cdf[:,None]), axis=0)
        n = len(self._obj_list)
        use_k[use_k==n] = n-1  # This shouldn't be possible, but maybe with numerical rounding...

        fluxPerPhoton = total_flux / len(photons)

        # Shoot the corresponding photons from each component.
        for kk, obj in enumerate(self.obj_list):
            use = (use_k == kk)  # True for each photon where this is the object to shoot from
            this_n = np.sum(use)
            if this_n == 0:
                continue
            temp = PhotonArray(this_n)
            temp._copyFrom(photons, slice(None), use, do_xy=False, do_flux=False)
            obj._shoot(temp, rng)
            temp.flux = fluxPerPhoton
            photons._copyFrom(temp, use, slice(None))

    def _drawKImage(self, image, jac=None):
        self.obj_list[0]._drawKImage(image, jac)
        if len(self.obj_list) > 1:
            im1 = image.copy()
            for obj in self.obj_list[1:]:
                obj._drawKImage(im1, jac)
                image += im1

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'FourierSqrt', 'FourierSqrtProfile' ]

import numpy as np
import copy

from .gsparams import GSParams
from .gsobject import GSObject
from .utilities import lazy_property
from .errors import galsim_warn
from . import chromatic as chrom


def FourierSqrt(obj, gsparams=None, propagate_gsparams=True):
    """A function for computing the Fourier-space square root of either a `GSObject` or
    `ChromaticObject`.

    The FourierSqrt function is principally used for doing an optimal coaddition algorithm
    originally developed by Nick Kaiser (but unpublished) and also described by Zackay & Ofek 2015
    (http://adsabs.harvard.edu/abs/2015arXiv151206879Z).  See the script make_coadd.py in the
    GalSim/examples directory for an example of how it works.

    This function will inspect its input argument to decide if a `FourierSqrtProfile` object or a
    `ChromaticFourierSqrtProfile` object is required to represent the operation applied to a surface
    brightness profile.

    Parameters:
        obj:                The object to compute the Fourier-space square root of.
        gsparams:           An optional `GSParams` argument. [default: None]
        propagate_gsparams: Whether to propagate gsparams to the transformed object.  This
                            is normally a good idea, but there may be use cases where one
                            would not want to do this. [default: True]

    Returns:
        a `FourierSqrtProfile` or `ChromaticFourierSqrtProfile` instance as appropriate.
    """

    if isinstance(obj, chrom.ChromaticObject):
        return chrom.ChromaticFourierSqrtProfile(obj, gsparams=gsparams,
                                                 propagate_gsparams=propagate_gsparams)
    elif isinstance(obj, GSObject):
        return FourierSqrtProfile(obj, gsparams=gsparams, propagate_gsparams=propagate_gsparams)
    else:
        raise TypeError("Argument to FourierSqrt must be either a GSObject or a ChromaticObject.")


class FourierSqrtProfile(GSObject):
    """A class for computing the Fourier-space sqrt of a `GSObject`.

    The FourierSqrtProfile class represents the Fourier-space square root of another profile.
    Note that the FourierSqrtProfile class, or compound objects (Sum, Convolution) that include a
    FourierSqrtProfile as one of the components cannot be photon-shot using the 'phot' method of
    `GSObject.drawImage` method.

    You may also specify a ``gsparams`` argument.  See the docstring for `GSParams` for more
    information about this option.  Note: if ``gsparams`` is unspecified (or None), then the
    FourierSqrtProfile instance inherits the same `GSParams` as the object being operated on.

    The normal way to use this class is to use the `FourierSqrt` factory function::

        >>> fourier_sqrt = galsim.FourierSqrt(obj)

    Parameters:
        obj:                The object to compute Fourier-space square root of.
        gsparams:           An optional `GSParams` argument. [default: None]
        propagate_gsparams: Whether to propagate gsparams to the transformed object.  This
                            is normally a good idea, but there may be use cases where one
                            would not want to do this. [default: True]
    """
    _sqrt2 = 1.4142135623730951

    _has_hard_edges = False
    _is_analytic_x = False

    def __init__(self, obj, gsparams=None, propagate_gsparams=True):
        if not isinstance(obj, GSObject):
            raise TypeError("Argument to FourierSqrtProfile must be a GSObject.")

        # Save the original object as an attribute, so it can be inspected later if necessary.
        self._gsparams = GSParams.check(gsparams, obj.gsparams)
        self._propagate_gsparams = propagate_gsparams
        if self._propagate_gsparams:
            self._orig_obj = obj.withGSParams(self._gsparams)
        else:
            self._orig_obj = obj

    @property
    def orig_obj(self):
        """The original object being Fourier sqrt-ed.
        """
        return self._orig_obj

    @property
    def _noise(self):
        if self.orig_obj.noise is not None:
            galsim_warn("Unable to propagate noise in galsim.FourierSqrtProfile")
        return None

    def withGSParams(self, gsparams=None, **kwargs):
        """Create a version of the current object with the given gsparams

        .. note::

            Unless you set ``propagate_gsparams=False``, this method will also update the gsparams
            of the wrapped component object.
        """
        if gsparams == self.gsparams: return self
        ret = copy.copy(self)
        ret._gsparams = GSParams.check(gsparams, self.gsparams, **kwargs)
        if self._propagate_gsparams:
            ret._orig_obj = self._orig_obj.withGSParams(ret._gsparams)
        return ret

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, FourierSqrtProfile) and
                 self.orig_obj == other.orig_obj and
                 self.gsparams == other.gsparams and
                 self._propagate_gsparams == other._propagate_gsparams))

    def __hash__(self):
        return hash(("galsim.FourierSqrtProfile", self.orig_obj, self.gsparams,
                     self._propagate_gsparams))

    def __repr__(self):
        return 'galsim.FourierSqrtProfile(%r, gsparams=%r, propagate_gsparams=%r)'%(
                self.orig_obj, self.gsparams, self._propagate_gsparams)

    def __str__(self):
        return 'galsim.FourierSqrt(%s)'%self.orig_obj

    def _prepareDraw(self):
        self.orig_obj._prepareDraw()

    @property
    def _maxk(self):
        return self.orig_obj.maxk

    @property
    def _stepk(self):
        return self.orig_obj.stepk * self._sqrt2

    @property
    def _is_axisymmetric(self):
        return self.orig_obj.is_axisymmetric

    @property
    def _is_analytic_k(self):
        return self.orig_obj.is_analytic_k

    @property
    def _centroid(self):
        return 0.5 * self.orig_obj.centroid

    @property
    def _flux(self):
        return np.sqrt(self.orig_obj.flux)

    @property
    def _positive_flux(self):
        return np.sqrt(self.orig_obj.positive_flux)

    @property
    def _negative_flux(self):
        return np.sqrt(self.orig_obj.negative_flux)

    @lazy_property
    def _flux_per_photon(self):
        return self._calculate_flux_per_photon()

    @property
    def _max_sb(self):
        # In this case, we want the autoconvolution of this object to get back to the
        # maxSB value of the original obj
        # flux * maxSB / 2 = maxSB_orig
        # maxSB = 2 * maxSB_orig / flux
        return 2. * self.orig_obj.max_sb / self.flux

    def _kValue(self, pos):
        return np.sqrt(self.orig_obj._kValue(pos))

    def _drawKImage(self, image, jac=None):
        self.orig_obj._drawKImage(image, jac)
        image.array[:,:] = np.sqrt(image.array)
